#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "picture_coder.h"
#include "vvl_defs.h"
#include "cpia.h"

extern int debug_flag;
extern char *frame_dir;


/* with thanks to Mads Dydensborg <madsdyd@diku.dk> for this routine */
/*
gstream_to_ppm save gstream format in memory as ppm format file.
*/
gstream_to_ppm(\
int xsize, int ysize, unsigned char *inbuffer, int inbuffer_size,\
FILE *fptr, int compression_flag)
{
int i;
int rgb_size;
/*
NOTE IMPORTANT:
This array has to be static, since in mpeg mode, in transmit, when
compression is on, data ACCUMULATES here (normally in the X display buffer)
*/
static unsigned char temp_buffer[382 * 288 * 4];

if(debug_flag)
	{
	fprintf(stderr, "gstream_to_ppm():\n\
	arg xsize=%d ysize=%d inbuffer=%lu inbuffer_size=%d\n\
	fptr=%lu compression_flag=%d\n",\
	xsize, ysize, (long)inbuffer, inbuffer_size,\
	(long)fptr, compression_flag);
	}

/* Allocate a buffer to save the picture in */
rgb_size = xsize * ysize * 3;

/* Convert it to rgb, 24 bit */
if(compression_flag)
	{
	if(! stream2rgb24c(inbuffer + 64, temp_buffer, inbuffer_size - 64) )
		{
		fprintf(stderr,\
		"gstream_to_ppm(): compressed, error converting to 24 bit, shit...\n");

		fclose(fptr);
		return 0;
		} 
	}
else/* not compressed */
	{
	if(! stream2rgb24(inbuffer + 64, temp_buffer, inbuffer_size - 64) )
		{
		fprintf(stderr,\
		"gstream_to_ppm(): uncompressed, error converting to 24 bit, shit...\n");

		fclose(fptr);
		return 0;
		} 
	}

/* write a ppm file. */
fprintf(fptr, "P6\n%i %i\n255\n", xsize, ysize);

/* Somehow, it seems that this is not rgb data but bgr data */
for(i = 0; i < rgb_size; i += 3)
	{
	fwrite(temp_buffer + i + 2, 1, 1, fptr);
	fwrite(temp_buffer + i + 1, 1, 1, fptr);
	fwrite(temp_buffer + i, 1, 1, fptr);
	}

fclose(fptr);

return 1;
}/* end function gstream_to_ppm */


/*
ppm_to_gstream  reads a ppm file and creates a gstream format in memory.
Copyright Jan Mourer
*/
int ppm_to_gstream(\
char *filename, BYTE __huge *out_buffer, int *bytes_in_buffer)
{
int a, c, i, j;
int r, g, b;
int width, height, maxval;
double y, u, v;
double cr, cg, cb, cu, cv;
BYTE __huge *out_ptr;
char temp[4096];
FILE *fptr;
int bytes_out;
int line_length;
int u_time;

if(debug_flag)
	{
	fprintf(stderr,\
	"ppm_to_gstream(): arg filename=%s out_buffer=%lu\n",\
	filename, out_buffer);
	}

if(! filename) return 0;
if(! out_buffer) return 0;
if(! bytes_in_buffer) return 0;

/* color standard */
/* Y spec */
cr = 0.3;
cg = 0.59;
cb = 0.11;

/* U spec */
cu = .5 / (1.0 - cb);

/* V spec */
cv = .5 / (1.0 - cr);

/* open the ppm format file */
sprintf(temp, "%s/mcam-jpeg.out", frame_dir);
fptr = fopen(temp, "rb");
if(!fptr)
	{
	fprintf(stderr,\
	"ppm_to_gstream(): jpeg cannot open ppm format file %s for read\n",\
	temp);
	strerror(errno);
	return 0;
	}

/* test for not closed files */
if(debug_flag) fprintf(stderr, "fileno(fptr)=%d\n", fileno(fptr) );

/*
cjpeg writes like this:
 "P6\n%ld %ld\n%d\n"
*/

/*
using fscanf here swallows control M after \n, resulting in wrong color,
perhaps a libc bug?
*/

/* parse the header */
i = 0;
j = 0;
while(1)
	{
	while(1)
		{
		errno = 0;
		c = getc(fptr);
		if(errno == EAGAIN) continue;
		if(errno == EINTR) continue;
		break;
		}
	if(c == EOF)
		{
		fclose(fptr);
		fprintf(stderr, "ppm_to_gstream(): early EOF in header\n");
		return 0;
		}
	temp[i] = c;
	if( (c == '\n') || (c == '\t') || (c == '\r') || (c == ' ') )
		{
		temp[i] = 0;
		i = 0;
		if(j == 0) {}; /* scip the P6 */
		if(j == 1) width = atoi(temp);
		if(j == 2) height = atoi(temp);
		if(j == 3) maxval = atoi(temp);
		j++;
		}/* end if white space */
	else i++;
	if(j == 4) break;	
	}/* end for all chars in header */

if(debug_flag)
	{
	fprintf(stderr,\
	"ppm_to_gstream(): width=%d height=%d maxval=%d\n",\
	width, height, maxval); 
	}

bytes_out = 0;
out_ptr = out_buffer;

/* gstream line length is 352 * 2 + EOL or 176 * 2 + EOL */
line_length = (width * 2) + 1; 

for (i = 0; i < height; i++)
	{
	/* gstream low byte line_length */
	*out_ptr = line_length % 256;
	out_ptr++;	

	/* gstream high byte line_lenght */
	*out_ptr = line_length / 256;
	out_ptr++;

	bytes_out += 2;
		
	/* write Y, U, Y, V */
	if(debug_flag)
		{
		fprintf(stderr, "ppm_to_gstream(): i=%d j=%d\n", i, j);
		}

	u_time = 1;
	for(j = 0; j < width; j++)
		{
		/* red */
		while(1)
			{
			errno = 0;
			r = getc(fptr);
			if(errno == EAGAIN) continue;
			if(errno == EINTR) continue;
			break;
			}

		if(r == EOF)
			{
			fprintf(stderr, "ppm_to_gstream(): early EOF r\n");
			/*
			this is a hack to fix early EOF in the output from cjpeg
			Substitute black.
			It causes the correct EOL to be written always, so the gstream
			format is not violated, as this results in decoding errors.
			*/
			r = 0;
/*			break;*/
			}

		/* green */
		while(1)
			{
			errno = 0;
			g = getc(fptr);
			if(errno == EAGAIN) continue;
			if(errno == EINTR) continue;
			break;
			}
		if(g == EOF)
			{
			fprintf(stderr, "ppm_to_gstream(): early EOF g\n");
			g = 0;
/*			break;*/
			}

		/* blue */
		while(1)
			{
			errno = 0;
			b = getc(fptr);
			if(errno == EAGAIN) continue;
			if(errno == EINTR) continue;
			break;
			}
		if(b == EOF)
			{
			fprintf(stderr, "ppm_to_gstream(): early EOF b\n");
			b = 0;
/*			break;*/
			}

		/* convert to YUV */

		/* test yuv coding here */
		y = cr * r + cg * g + cb * b;

		y = (219.0 / 256.0) * y + 16.5;  /* nominal range: 16..235 */

		*out_ptr = y;
		out_ptr++;

		if(u_time)
			{
			u = cu * (b - y);
			u = (224.0 / 256.0) * u + 128.5; /* 16..240 */

			*out_ptr = u;
			}

		else
			{
			v = cv * (r - y);
			v = (224.0 / 256.0) * v + 128.5; /* 16..240 */

			*out_ptr = v;
			}
			
		out_ptr++;
		u_time = 1 - u_time;
		
		bytes_out += 4;
		}/* end for j horizontal */

	/* write gstream EOL marker */
	*out_ptr = 0xfd;
	out_ptr++;

	bytes_out += 1;
	}/* end for i vertical */

/* close the input file */
fclose(fptr);

/* write gstream EOF markers */
*out_ptr = 0xff;
out_ptr++;
*out_ptr = 0xff;
out_ptr++;
*out_ptr = 0xff;
out_ptr++;
*out_ptr = 0xff;

bytes_out += 4;

if(debug_flag)
	{
	fprintf(stderr, "ppm_to_gstream(): bytes_out=%d\n", bytes_out);
	}

*bytes_in_buffer = bytes_out;

return 1;
}/* end function ppm_to_gstream */ 


/*
gstream_to_yuv save gstream format in memory as ppm format file.
*/
gstream_to_yuv(int xsize, int ysize, unsigned char *inbuffer, int inbuffer_size, int compression_flag)
{
int a, i, j;
static int have_header = 0;
int yuv_buffer_length;
char yuv_buffer[ (352 * 288) + ( (352 / 2) * (288 / 2) ) + ( (352 / 2) * (288 / 2) ) ];
int r, g, b;
double y, u, v;
double cr, cg, cb, cu, cv;
int u_time;
unsigned char *y_ptr, *u_ptr, *v_ptr;
int odd_line;
unsigned char *t_ptr;

/*
NOTE IMPORTANT:
This array has to be static, since in mpeg mode, in transmit, when
compression is on, data ACCUMULATES here (normally in the X display buffer)
*/
static unsigned char temp_buffer[382 * 288 * 4];

if(debug_flag)
	{
	fprintf(stderr, "gstream_to_ppm():\n\
	arg xsize=%d ysize=%d inbuffer=%p inbuffer_size=%d compression_flag=%d\n",\
	xsize, ysize, inbuffer, inbuffer_size, compression_flag);
	}

/* color standard */
/* Y spec */
cr = 0.3;
cg = 0.59;
cb = 0.11;

/* U spec */
cu = .5 / (1.0 - cb);

/* V spec */
cv = .5 / (1.0 - cr);

/* write yuv header to stderr */
if(! have_header)
	{
	// estimate camera speed 1 fps
	fprintf(stdout, "YUV4MPEG2 W%d H%d F%d:1 I? A0:0 XM2AR002\n", xsize, ysize, 1);

	have_header = 1;
	}

/* Convert it to rgb, 24 bit */
if(compression_flag)
	{
	if(! stream2rgb24c(inbuffer + 64, temp_buffer, inbuffer_size - 64) )
		{
		fprintf(stderr,\
		"gstream_to_yuv(): compressed, error converting to 24 bit, shit...\n");

		return 0;
		} 
	}
else/* not compressed */
	{
	if(! stream2rgb24(inbuffer + 64, temp_buffer, inbuffer_size - 64) )
		{
		fprintf(stderr,\
		"gstream_to_yuv(): uncompressed, error converting to 24 bit, shit...\n");

		return 0;
		} 
	}

/* output frame header */
fprintf(stdout, "FRAME\n");

/* convert rgb to yuv */

/* for RGB to YUV */
y_ptr = yuv_buffer;
u_ptr = yuv_buffer + (xsize * ysize);
v_ptr = u_ptr + ( (xsize / 2) * (ysize / 2) );

t_ptr = temp_buffer;

/* from temp_buffer in rgb 24 format to yuv_buffer */
odd_line = 0;
for(i = 0; i < ysize; i++)
	{
	u_time = 1;
	for(j = 0; j < xsize; j++)
		{
		/* convert to YUV */

		b = (int) *t_ptr;
		t_ptr++;
		g = (int) *t_ptr;
		t_ptr++;
		r = (int) *t_ptr;
		t_ptr++;		

		/* test yuv coding here */
		y = cr * r + cg * g + cb * b;
//		y = (219.0 / 256.0) * y + 16.5;  /* nominal range: 16..235 */

		*y_ptr = y;
		y_ptr++;

		if(odd_line)
			{
			if(u_time)
				{
				u = cu * (b - y);

//				u = (224.0 / 256.0) * u + 128.5; /* 16..240 */
//				*u_ptr = u;				
		
				*u_ptr = u + 128;
				u_ptr++;
				}
			else
				{
				v = cv * (r - y);

//				v = (224.0 / 256.0) * v + 128.5; /* 16..240 */
//				*v_ptr = v;

				*v_ptr = v + 128;
				v_ptr++;
				}
			} /* end if odd_line */
	
		u_time = 1 - u_time;
		} /* end if j */

	odd_line = 1 - odd_line;
	} /* end for i */

/* write the yuv data */
yuv_buffer_length = (xsize * ysize) + ( (xsize / 2) * (ysize / 2) ) + ( (xsize / 2) * (ysize / 2) );

if(debug_flag)
	{
	fprintf(stderr, "yuv_buffer_length=%d", yuv_buffer_length);
	}

a = fwrite(yuv_buffer, sizeof(char), yuv_buffer_length, stdout); 
if(a != yuv_buffer_length)
	{
	fprintf(stderr, "gstream_to_yuv(): could only write %d of %d bytes, aborting.\n",\
	a, yuv_buffer_length);

	exit(1);
	}

return 1;
} /* end function gstream_to_yuv */


